// The MIT License
//
// Copyright (c) 2003 Ron Alford, Mike Grove, Bijan Parsia, Evren Sirin
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to
// deal in the Software without restriction, including without limitation the
// rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
// sell copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
// FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
// IN THE SOFTWARE.


package org.mindswap.pellet.utils;


import aterm.*;
import aterm.pure.*;
import org.mindswap.pellet.*;

import java.util.*;


/**
 * This class provides the functions ATerm related functions. Creating terms for URI's
 * and complex class descriptions is done here. There are also functions for normalization, 
 * simplification and conversion to NNF (Normal Negation Form).  
 * 
 * @author initial version for OWL - Evren Sirin, updates for OWLMeT - NataKeberle
 */
public class ATermUtils {
	// concept constructors
    //static int i=0;
	public static final String NOT  = "not";
	public static final String AND  = "and";
	public static final String OR   = "or";
	public static final String SOME = "some";
	public static final String ALL  = "all";
	public static final String MIN  = "min";
	public static final String MAX  = "max";
	public static final String CARD = "card";
	public static final String VALUE= "value";
	
	// role
	public static final String INV  = "inv";
	
	//tbox
	public static final String SUB  = "sub";
	public static final String SAME = "same";
	//for owlmet
	public static final String FUT = "fut";
	public static final String PAST = "past";
	public static final String AT = "at";
	public static final String ALLFUT = "allfut";
	public static final String ALLPAST = "allpast";
	public static final String SOMEFUT = "somefut";
	public static final String SOMEPAST = "somepast";
	
	private static final ATermFactory factory = new PureFactory();
	
	public static final AFun LITFUN  = factory.makeAFun("literal", 3, false);
	public static final int LIT_VAL_INDEX  = 0;
	public static final int LIT_LANG_INDEX = 1;
	public static final int LIT_URI_INDEX  = 2;

    public static final AFun ANDFUN  = factory.makeAFun(AND, 1, false);
    public static final AFun ORFUN   = factory.makeAFun(OR, 1, false);
    public static final AFun SOMEFUN = factory.makeAFun(SOME, 2, false);
    public static final AFun ALLFUN  = factory.makeAFun(ALL, 2, false);
    public static final AFun NOTFUN  = factory.makeAFun(NOT, 1, false);
	public static final AFun MAXFUN  = factory.makeAFun(MAX, 2, false);
	public static final AFun MINFUN  = factory.makeAFun(MIN, 2, false);
	public static final AFun VALUEFUN= factory.makeAFun(VALUE, 1, false);

	public static final AFun FUTFUN  = factory.makeAFun(FUT, 2, false);
	public static final AFun PASTFUN  = factory.makeAFun(PAST, 2, false);
	public static final AFun ATFUN  = factory.makeAFun(AT, 2, false);
	public static final AFun ALLFUTFUN  = factory.makeAFun(ALLFUT, 1, false);
	public static final AFun ALLPASTFUN  = factory.makeAFun(ALLPAST, 1, false);
	public static final AFun SOMEFUTFUN  = factory.makeAFun(SOMEFUT, 1, false);
	public static final AFun SOMEPASTFUN  = factory.makeAFun(SOMEPAST, 1, false);
	
	public static Set CLASS_FUN = SetUtils.create( new AFun[] { 
	    ALLFUN, SOMEFUN, MAXFUN, MINFUN, ANDFUN, ORFUN, NOTFUN, VALUEFUN ,
		FUTFUN, PASTFUN, ATFUN, ALLFUTFUN, ALLPASTFUN, SOMEFUTFUN, SOMEPASTFUN
	} );

	/**
	 * This is not used in the reasoner but kept here to be used for display
	 */
	public static final AFun CARDFUN  = factory.makeAFun(CARD, 2, false);
    
    public static final AFun INVFUN  = factory.makeAFun(INV, 1, false);
	public static final AFun SUBFUN  = factory.makeAFun(SUB, 2, false);
	public static final AFun SAMEFUN = factory.makeAFun(SAME, 2, false);
	
	/**
	 * This is used to represent variables in queries
	 */
	public static final AFun VARFUN  = factory.makeAFun("var", 1, false);
    

	public static final AFun TYPEFUN  = factory.makeAFun("type", 2, false);
	public static final AFun IPFUN  = factory.makeAFun("iptriple", 3, false);
	public static final AFun DPFUN = factory.makeAFun("dptriple", 3, false);
	
	public static final ATermAppl EMPTY = makeTermAppl( "" );
	
	public static final ATermList EMPTY_LIST = factory.makeList();

	// TOP and BOTTOM concepts. TOP is not defined as T or not(T) any 
	// more but added to each node manually. Defining TOP as a primitive
	// concept reduces number of GCIs and makes other reasoning tasks 
	// faster
	public static final ATermAppl TOP    = ATermUtils.makeTermAppl("_TOP_");
	public static final ATermAppl BOTTOM = ATermUtils.makeNot(TOP);
	
    public static final ATermInt ONE = factory.makeInt( 1 );	
	
	static public ATermFactory getFactory() {
		return factory;	
	}			
	
	static public ATermAppl makePlainLiteral(String value) {
		return factory.makeAppl(ATermUtils.LITFUN, 
			makeTermAppl(value), 
			EMPTY, 
			EMPTY);
	}
	
	static public ATermAppl makePlainLiteral(String value, String lang) {
		return factory.makeAppl(ATermUtils.LITFUN, 
			makeTermAppl(value), 
			makeTermAppl(lang), 
			EMPTY);
	}
	
	static public ATermAppl makeTypedLiteral(String value, String dt) {
		return factory.makeAppl(ATermUtils.LITFUN, 
			makeTermAppl(value), 
			EMPTY, 
			makeTermAppl(dt));
	}
    
    static public ATermAppl NO_DATATYPE = makeTermAppl("NO_DATATYPE");
    static public ATermAppl makeLiteral(ATermAppl name) {
        return factory.makeAppl(ATermUtils.LITFUN, 
            name, 
            EMPTY, 
            NO_DATATYPE);
    }

	static public ATermAppl makeTermAppl(String name) {
		return factory.makeAppl(factory.makeAFun(name, 0 , false));	
	}

	static public ATermAppl makeTermAppl(AFun fun, ATerm[] args) {
		return factory.makeAppl(fun, args);	
	}
	
	static public ATermAppl makeNot(ATerm c) {
		return factory.makeAppl(NOTFUN, c);	
	}			

	static public ATerm term(String str) {
		return factory.parse(str);	
	}
	
	// negate all the elements in the list and return the new list	
	static public ATermList negate(ATermList list) {		
		if(list.isEmpty())
			return list;

		ATermAppl a = (ATermAppl) list.getFirst();
		a = isNot(a) ? (ATermAppl) a.getArgument(0): makeNot(a);
		ATermList result = makeList(a, negate(list.getNext()));
		
		return result;
	}

	final static public ATermAppl negate(ATermAppl a) {
		return isNot(a) ? (ATermAppl) a.getArgument(0) : makeNot(a);
	}
	
	final static public ATermAppl makeVar(String name) {
		return factory.makeAppl(VARFUN, makeTermAppl(name));	
	}	
	
	final static public ATermAppl makeVar(ATermAppl name) {
		return factory.makeAppl(VARFUN, name);	
	}	
			
	final static public ATermAppl makeValue(ATerm c) {
		return factory.makeAppl(VALUEFUN, c);	
	}			

	final static public ATermAppl makeInv(ATermAppl r) {
		return factory.makeAppl(INVFUN, r);	
	}			

	final static public ATermAppl makeSub(ATerm a, ATerm b) {
		return factory.makeAppl(SUBFUN, a, b);	
	}			

	final static public ATermAppl makeSame(ATerm a, ATerm b) {
		return factory.makeAppl(SAMEFUN, a, b);	
	}			

	final static public ATermAppl makeAnd(ATerm c1, ATerm c2) {
		return makeAnd(makeList(c2).insert(c1));	
	}		

    static public ATermAppl makeAnd(ATermList list) {
    	if(list == null || list.isEmpty())
    		return BOTTOM;
    	else if(list.getNext().isEmpty())
    		return (ATermAppl) list.getFirst();
    		
		return factory.makeAppl(ANDFUN, list);	
	}		

    final static public ATermAppl makeOr(ATermAppl c1, ATermAppl c2) {
		return makeOr(makeList(c2).insert(c1));	
	}		

	static public ATermAppl makeOr(ATermList list) {
		if(list == null || list.isEmpty())
			return BOTTOM;
		else if(list.getNext().isEmpty())
			return (ATermAppl) list.getFirst();
    		
		return factory.makeAppl(ORFUN, list);	
	}			

	final static public ATermAppl makeAllValues(ATerm r, ATerm c) {
		return factory.makeAppl(ALLFUN, r, c);	
	}			

	final static public ATermAppl makeSomeValues(ATerm r, ATerm c) {
		assertTrue( c instanceof ATermAppl );
		
		return factory.makeAppl(SOMEFUN, r, c);	
	}			

	final static public ATermAppl makeHasValue(ATerm r, ATerm ind) {
	    ATermAppl c = makeValue( ind );
		return factory.makeAppl(SOMEFUN, r, c);	
	}			

	final static public ATermAppl makeDisplayCard(ATerm r, int n) {
	    assertTrue(n >= 0);
        
	    return factory.makeAppl(CARDFUN, r, factory.makeInt(n));
	}

	final static public ATermAppl makeNormalizedMax(ATerm r, int n) {
		assertTrue(n >= 0);

		return makeNot(makeMin(r, n + 1));
	}
	
	final static public ATermAppl makeMax(ATerm r, int n) {
		assertTrue(n >= 0);

		// This was causing nnf to come out wrong
		//return makeNot(makeMin(r, n + 1));
		
		return factory.makeAppl(MAXFUN, r, factory.makeInt(n));	
	}	
			
	final static public ATermAppl makeMin(ATerm r, int n) {
		if(n == 0) return ATermUtils.TOP;
		
		assertTrue(n > 0);

		return factory.makeAppl(MINFUN, r, factory.makeInt(n));	
	}	
	
	
	final static public ATermAppl makeDisplayMax(ATerm r, int n) {
		assertTrue(n >= 0);

		return factory.makeAppl(MAXFUN, r, factory.makeInt(n));
	}	
			
	final static public ATermAppl makeDisplayMin(ATerm r, int n) {
		assertTrue(n >= 0);

		return factory.makeAppl(MINFUN, r, factory.makeInt(n));	
	}
	
	final static public ATermAppl makeCard(ATerm r, int n) {
		ATermAppl max = makeMax(r, n);
        if(n == 0) return max;
        
        ATermAppl min = makeMin(r, n);
        return makeAnd(min, max);
	}	
	
	final static public ATermList makeList(ATerm singleton) {
		return factory.makeList(singleton, EMPTY_LIST);	
	}		

	final static public ATermList makeList(ATerm first, ATermList rest) {
		return factory.makeList(first, rest);	
	}		
    
    public static ATermList makeList(Collection set) {
		ATermList list = EMPTY_LIST;
	
		for (Iterator iter = set.iterator(); iter.hasNext();) {
		    list = list.insert((ATerm)iter.next());
		}
		return list;
    }
    
	final static public ATermList makeList(ATerm[] aTerms) {		
		return makeList(aTerms, 0);	
	}		

	static private ATermList makeList(ATerm[] aTerms, int index) {		
		if(index >= aTerms.length)
			return EMPTY_LIST;
		else if(index == aTerms.length - 1)
			return makeList(aTerms[index]);
				
		return makeList(aTerms[index], makeList(aTerms, index+1));	
	}		

//Added for owlmet
	final static public ATermAppl makeFuture(ATerm r, int n) {
		//if(n == 0) return ATermUtils.TOP;
		
		assertTrue(n >= 0);

		return factory.makeAppl(FUTFUN, r, factory.makeInt(n));	
	}

	final static public ATermAppl makePast(ATerm r, int n) {
		//if(n == 0) return ATermUtils.TOP;
		
		assertTrue(n >= 0);

		return factory.makeAppl(PASTFUN, r, factory.makeInt(n));	
	}

	final static public ATermAppl makeAt(ATerm a, ATerm b) {
		return factory.makeAppl(ATFUN, a, b);	
	}
	
	final static public ATermAppl makeAllfuture(ATerm a) {
		return factory.makeAppl(ALLFUTFUN, a);	
	}
	
	final static public ATermAppl makeAllpast(ATerm a) {
		return factory.makeAppl(ALLPASTFUN, a);	
	}
	
	final static public ATermAppl makeSomefuture(ATerm a) {
		return factory.makeAppl(SOMEFUTFUN, a);	
	}
	
	final static public ATermAppl makeSomepast(ATerm a) {
		return factory.makeAppl(SOMEPASTFUN, a);	
	}
    //end of added for owlmet

	final static public boolean member(ATerm a, ATermList list) {
		return (list.indexOf(a, 0) != -1);
	}
	
	static public boolean isSet(ATermList list) {
		if(list.isEmpty())
			return true;

		ATerm curr = list.getFirst();
		list = list.getNext();
		while(!list.isEmpty()) {
			ATerm next = list.getFirst();
			if(Comparators.hashCodeComparator.compare(curr, next) >= 0)
				return false;
			curr = next;	
			list = list.getNext();	
		} 
		
		return true;
	}
	
	static public ATermList toSet(ATermList list) {
		if(isSet(list))
			return list;
			
		int size = list.getLength();	
				
		ATerm[] a = toArray(list);
		if(a == null || a.length < size)	
			a = new ATerm[Math.max(100, size)];

		Arrays.sort(a, 0, size, Comparators.hashCodeComparator);
				
		ATermList set = makeList(a[size - 1]);
		for(int i = size - 2; i >= 0; i--) {
			ATerm s = set.getFirst();
			if(!s.equals(a[i]))
				set = set.insert(a[i]);
		}
			
		return set;	
	}
	
	static public ATermList toSet(ATerm[] a, int size) {
		Arrays.sort(a, 0, size, Comparators.hashCodeComparator);
				
		ATermList set = makeList(a[size - 1]);
		for(int i = size - 2; i >= 0; i--) {
			ATerm s = set.getFirst();
			if(!s.equals(a[i]))
				set = set.insert(a[i]);
		}
			
		return set;	
	}

	static public ATermList toSet( Collection coll ) {
		int size = coll.size();	
				
		ATerm[] a = new ATermAppl[ size ];
		coll.toArray( a );

		return toSet( a, size );	
	}
	
	public static String toString(ATermAppl term) {
       if( isVar( term ) )
           return ((ATermAppl)term.getArgument(0)).getName();
       else if( ATermUtils.isLiteral( term ) ) {		
           String value = ((ATermAppl) term.getArgument(0)).getName();
           String lang = ((ATermAppl) term.getArgument(1)).getName();
           String datatypeURI = ((ATermAppl) term.getArgument(2)).getName();
           
           StringBuffer sb = new StringBuffer();
           sb.append('"').append(value).append('"');
           if(!lang.equals("")) 
               sb.append('@').append(lang);
           else if(!datatypeURI.equals(""))
               sb.append("^^").append(datatypeURI);
                   
           return sb.toString();           
       }
       else
           return term.getName();
	}
	
	static public ATerm[] toArray(ATermList list) {
		ATerm[] a = new ATerm[list.getLength()];
		
		for(int i = 0; !list.isEmpty(); list = list.getNext())
			a[i++] = list.getFirst();
		
		return a;
	}

	public final static void assertTrue(boolean condition) {
		if(!condition) {
			throw new RuntimeException("assertion failed.");
		}
	}		

	public static final boolean isPrimitive(ATermAppl c) {
		return c.getArity() == 0;		
	}

	public static final boolean isNegatedPrimitive(ATermAppl c) {
		return isNot(c) && isPrimitive((ATermAppl) c.getArgument(0));		
	}
	
	public static final boolean isPrimitiveOrNegated(ATermAppl c) {
		return isPrimitive(c) || isNegatedPrimitive(c);		
	}
	
	public static final boolean isBnode( ATermAppl name ) {
		return name.getName().startsWith(PelletOptions.BNODE);
	}

	public static final boolean isAnon( ATermAppl name ) {
		return name.getName().startsWith(PelletOptions.ANON);
	}

	public static Set listToSet(ATermList list) {
		Set set = new HashSet();
		while(!list.isEmpty()) {
			set.add(list.getFirst());
			list = list.getNext();
		}
		return set;
	}

	public static Set getPrimitives(ATermList list) {
		Set set = new HashSet();
		while(!list.isEmpty()) {
		    ATermAppl term = (ATermAppl) list.getFirst();
		    if(isPrimitive(term))
		        set.add(term);
			list = list.getNext();
		}
		return set;
	}
    
	public final static boolean isAnd(ATermAppl a) {
		return a.getAFun().equals(ANDFUN);
	}

	public final static boolean isOr(ATermAppl a) {
		return a.getAFun().equals(ORFUN);
	}
	
	public final static boolean isAllValues(ATermAppl a) {
		return a.getAFun().equals(ALLFUN);
	}

	public final static boolean isSomeValues(ATermAppl a) {
		return a.getAFun().equals(SOMEFUN);
	}
	
	public final static boolean isHasValue(ATermAppl a) {
		return a.getAFun().equals(SOMEFUN) && ((ATermAppl) a.getArgument(1)).getAFun().equals(VALUEFUN);
	}
	
	public final static boolean isNominal(ATermAppl a) {
		return a.getAFun().equals(VALUEFUN);
	}
	
	public final static boolean isOneOf(ATermAppl a) {
		if(!a.getAFun().equals(ORFUN))
			 return false;
		
		ATermList list = (ATermList) a.getArgument(0);
		while(!list.isEmpty()) {
			if(!isNominal((ATermAppl) list.getFirst()))
				return false;
			list = list.getNext();
		}
		return true;
	}
	
	public final static boolean isDataRange(ATermAppl a) {
		if(!a.getAFun().equals(ORFUN))
			 return false;
		
		ATermList list = (ATermList) a.getArgument(0);
		while(!list.isEmpty()) {
		    ATermAppl term = (ATermAppl) list.getFirst();
			if(!isNominal(term) || !isLiteral((ATermAppl) term.getArgument(0)))
				return false;
			list = list.getNext();
		}
		return true;
	}
	
	public final static boolean isNot(ATermAppl a) {
		return a.getAFun().equals(NOTFUN);
	}
	
	public final static boolean isMax(ATermAppl a) {
		return a.getAFun().equals(MAXFUN);
	}
	
	public final static boolean isMin(ATermAppl a) {
		return a.getAFun().equals(MINFUN);
	}

	public final static boolean isCard(ATermAppl a) {
	    if( isMin(a) || isMax(a) )
	        return true;
	    else if( isAnd(a) ) {
	        a = (ATermAppl) a.getArgument(0);
	        return isMin(a) || isMax(a);
	    }
	    
		return false;
	}

	public final static boolean isLiteral(ATermAppl a) {
		return a.getAFun().equals(LITFUN);
	}
	
	//for owlmet
	public final static boolean isFuture(ATermAppl a) {
		return a.getAFun().equals(FUTFUN);
	}
	
	public final static boolean isPast(ATermAppl a) {
		return a.getAFun().equals(PASTFUN);
	}

	public final static boolean isAt(ATermAppl a) {
		return a.getAFun().equals(ATFUN);
	}
	
	public final static boolean isAllfuture(ATermAppl a) {
		return a.getAFun().equals(ALLFUTFUN);
	}
	
	public final static boolean isAllpast(ATermAppl a) {
		return a.getAFun().equals(ALLPASTFUN);
	}
	
	public final static boolean isSomefuture(ATermAppl a) {
		return a.getAFun().equals(SOMEFUTFUN);
	}
	
	public final static boolean isSomepast(ATermAppl a) {
		return a.getAFun().equals(SOMEPASTFUN);
	}//for owlmet
	
	final static public boolean isVar(ATermAppl a) {
		return a.getAFun().equals(VARFUN);
	}
    
    public static boolean isComplexClass(ATerm c) {
    	if (c instanceof ATermAppl) {
    		ATermAppl a = (ATermAppl) c;
    		AFun f = a.getAFun();
    		return CLASS_FUN.contains( f );
    	}
    	return false;
    }

    public static boolean isTimeStructure(ATermAppl c) {
        if (c.getName().equals("TimeStructure"))
            return true;
        else
            return false;
    }

	public static ATerm nnf(ATerm term) {
		if(term instanceof ATermList)
			return nnf((ATermList) term);
		if(term instanceof ATermAppl)
			return nnf((ATermAppl) term);

		return null;
	}
	
	public static ATermList nnf(ATermList list) {
		ATermList newList = factory.makeList();
		while (!list.isEmpty()) {
			newList = newList.append(nnf((ATermAppl) list.getFirst()));
			list = list.getNext();
		}

		return newList;
	}
	
    /*
     * return the term in NNF form, i.e. negation only occurs in
     * front of atomic concepts
     */
	public static ATermAppl nnf(ATermAppl term) {
		ATermAppl newterm = null;
		
//		if(new Throwable().fillInStackTrace().getStackTrace().length > 200)
//		    throw new InternalReasonerException("DEBUG: Too many nnf calls");
		
		AFun af = term.getAFun();

		if (af.equals(ATermUtils.NOTFUN)) { // Function is a NOT
			// Take the first argument to the NOT, then check
			// the type of that argument to determine what needs to be done.
			ATermUtils.assertTrue(af.getArity() == 1);
			ATermAppl arg = (ATermAppl) term.getArgument(0);
			af = arg.getAFun();
	
			if (arg.getArity() == 0) {
				newterm = term; // Negation is in as far as it can go
			}
			else if (af.equals(ATermUtils.NOTFUN)) { // Double negation.
				newterm = nnf((ATermAppl) arg.getArgument(0));
			}
			else if (af.equals(ATermUtils.VALUEFUN)) {
				newterm = term;
			}
			else if (af.equals(ATermUtils.MAXFUN)) {
				ATermInt n = (ATermInt) arg.getArgument(1);
				newterm = ATermUtils.makeMin(arg.getArgument(0), n.getInt() + 1);
			}
			else if (af.equals(ATermUtils.MINFUN)) {
				ATermInt n = (ATermInt) arg.getArgument(1);
				if (n.getInt() == 0)
					newterm = ATermUtils.BOTTOM;
				else
					newterm = ATermUtils.makeMax(arg.getArgument(0), n.getInt() - 1);
			}
			else if (af.equals(ATermUtils.ANDFUN)) {
				return ATermUtils.makeOr(nnf(negate((ATermList)arg.getArgument(0))));
			}
			else if (af.equals(ATermUtils.ORFUN)) {
				return ATermUtils.makeAnd(nnf(negate((ATermList)arg.getArgument(0))));
			}
			else if (af.equals(ATermUtils.SOMEFUN)) {
				newterm = ATermUtils.makeAllValues(arg.getArgument(0), nnf(makeNot(arg.getArgument(1))));
			}
			else if (af.equals(ATermUtils.ALLFUN)) {
				newterm = ATermUtils.makeSomeValues(arg.getArgument(0), nnf(makeNot(arg.getArgument(1))));
			}
			else if(af.equals(ATermUtils.ATFUN)){
				newterm = ATermUtils.makeAt(nnf( (ATerm) negate( (ATermAppl) arg.getArgument(0))), (ATerm) arg.getArgument(1));
			}
			else if(af.equals(ATermUtils.FUTFUN)){
				ATermInt n = (ATermInt) arg.getArgument(1);
				newterm = ATermUtils.makeFuture(nnf( (ATerm) negate( (ATermAppl) arg.getArgument(0))), n.getInt());
			}
			else if(af.equals(ATermUtils.PASTFUN)){
				ATermInt n = (ATermInt) arg.getArgument(1);
				newterm = ATermUtils.makePast(nnf( (ATerm) negate( (ATermAppl) arg.getArgument(0))), n.getInt());
			}
            else if (af.equals(ATermUtils.SOMEFUTFUN)){
                newterm = ATermUtils.makeAllfuture(nnf( (ATerm) negate ((ATermAppl) arg.getArgument(0))));
            }
            else if (af.equals(ATermUtils.SOMEPASTFUN)){
                newterm = ATermUtils.makeAllpast(nnf( (ATerm) negate ((ATermAppl) arg.getArgument(0))));
            }
            else if (af.equals(ATermUtils.ALLFUTFUN)){
            newterm = ATermUtils.makeAllfuture(nnf( (ATerm) negate ((ATermAppl) arg.getArgument(0))));
            }
            else if (af.equals(ATermUtils.ALLPASTFUN)){
                newterm = ATermUtils.makeAllpast(nnf( (ATerm) negate ((ATermAppl) arg.getArgument(0))));
            }
			else {
				ATermUtils.assertTrue(false);
			}
		}
		else if (af.equals(ATermUtils.MINFUN) || af.equals(ATermUtils.MAXFUN) || af.equals(ATermUtils.FUTFUN)
                || af.equals(ATermUtils.PASTFUN) ){
			return term;
		}
		else {
			// Return the term with all of its arguments in nnf
                ATerm args[] = new ATerm[term.getArity()];
			    for (int i = 0; i < term.getArity(); i++) {
                args[i] = nnf(term.getArgument(i));
    			}
			    newterm = factory.makeAppl(af, args);
		}
		ATermUtils.assertTrue(newterm != null);

		return newterm;
	} 

	public static Collection normalize( Collection coll ) {
	    List list = new ArrayList();
	    for( Iterator i = coll.iterator(); i.hasNext(); ) {
            ATermAppl term = (ATermAppl) i.next();
            list.add( normalize( term ) );
        }
		
		return list;
	}
	
	public static ATermList normalize( ATermList list ) {
	    int size = list.getLength();
		ATerm[] terms = new ATerm[ size ];
		for(int i = 0; i < size; i++) {
            terms[i] = normalize( (ATermAppl) list.getFirst() );
            terms[i] = normalize((ATermAppl)terms[i]);
            list = list.getNext();
        }

		ATermList set = toSet( terms, size );
		
		return set;
	}

	/**
	 * Normalize the term by making following changes:
	 * <ul>
	 *   <li>or([a1, a2,..., an]) -> not(and[not(a1), not(a2), ..., not(an)]])</li>
	 *   <li>some(p, c) -> not(all(p, not(c)))</li>
	 *   <li>max(p, n) -> not(min(p, n+1))</li>
	 * </ul>
	 * 
	 * @param term
	 * @return
	 */
	public static ATermAppl normalize(ATermAppl term) {
		ATermAppl norm = term;
		AFun      fun  = term.getAFun();
		ATerm     arg1  = (term.getArity() > 0) ? term.getArgument(0) : null;
		ATerm     arg2  = (term.getArity() > 1) ? term.getArgument(1) : null;

		if(fun.equals(NOTFUN)) { 
		    if( !isPrimitive( (ATermAppl) arg1 ) )
		        norm = simplify(makeNot(normalize((ATermAppl) arg1)));
		}
		else if(fun.equals(ANDFUN)) { 
			norm = simplify(makeAnd(normalize((ATermList) arg1)));
		}
		else if(fun.equals(ORFUN)) { 
			ATermList neg = negate((ATermList) arg1);
			ATermAppl and = makeAnd(neg);
			ATermAppl notAnd = makeNot(and);
			norm = normalize(notAnd);
		}
		else if(fun.equals(ALLFUN)) { 
			norm = simplify(makeAllValues(arg1, normalize((ATermAppl) arg2)));
		}
		else if(fun.equals(SOMEFUN)) { 
			norm = normalize(makeNot(makeAllValues(arg1, makeNot(arg2))));
		}
		else if(fun.equals(MAXFUN)) {
			norm = makeNot(makeMin(arg1, ((ATermInt) arg2).getInt() + 1));
		}
		else if(fun.equals(MINFUN)) {
			norm = simplify(norm);
		}
		else if(fun.equals(FUTFUN) || fun.equals(PASTFUN)){
			ATerm arg11 = (((ATermAppl) arg1).getArity() > 0) ? ((ATermAppl) arg1).getArgument(0) : null;
			ATerm arg12 = (((ATermAppl) arg1).getArity() > 1) ? ((ATermAppl) arg1).getArgument(1) : null;
			AFun arg1fun = ((ATermAppl) arg1).getAFun();
			
            if (arg1fun.equals(NOTFUN) && ((ATermAppl)arg11).getArity()>=1){
                  norm = factory.makeAppl(fun, nnf((ATermAppl)arg1),arg2);
                  norm = normalize(norm);
            } else
            if(arg1fun.equals(FUTFUN) || arg1fun.equals(PASTFUN)){ //theorems 5, 6, 7, 8, PTC(MT) paper
				ATermInt n1 = (ATermInt) arg2;
				int num1 = n1.getInt();
				if(fun.equals(PASTFUN)){
					num1*=-1;
				}
				ATermInt n2 = (ATermInt) arg12;
				int num2 = n2.getInt();
				if(arg1fun.equals(PASTFUN)){
					num2*=-1;
				}
				int num = num1 + num2;
				
				if(num>0){
					norm = normalize(makeFuture(arg11,num));
				}
				else if(num<0){
					norm = normalize(makePast(arg11,-num));
				}
				else norm = normalize((ATermAppl)arg11);
			}
//TODO S4 modal system: somepsomep = somep, allpallp = allp,non-reducible modalities are 12: allp, allpsomep, allpsomepallp, somep, somepallp, somepallpsomep and their pairs with "not"
			else if(arg1fun.equals(ALLFUTFUN) || arg1fun.equals(ALLPASTFUN) ||
						arg1fun.equals(SOMEFUTFUN) || arg1fun.equals(SOMEPASTFUN)){ //theorems 10, 11, 12, 13, 14, 15, 16, 17 [prop.metric.temporal logic paper]
//TODO  exchange allfut(fut(a,num)) with fut(allfut(a),num)) as we need fut to be first modality in tableau expansion
				    norm = factory.makeAppl(fun,normalize(factory.makeAppl(arg1fun,arg11)),arg2);
            }
			else if(arg1fun.equals(ATFUN)){//theorems 30, 31 [hybrid prop. metric temporal logic paper]
				norm = normalize(makeAt( arg11, arg12 ));
			}
			else if(((ATermInt)arg2).getInt()==0){
                norm = normalize((ATermAppl)arg1);
			}
		}
//TODO non-reducible are: allf, allfsomef, allfsomefallf, somef, somefallf, somefallfsomef and their pairs with "not"
		else if(fun.equals(SOMEFUTFUN) || fun.equals(SOMEPASTFUN)){
			ATerm arg11 = (((ATermAppl) arg1).getArity() > 0) ? ((ATermAppl) arg1).getArgument(0) : null;
			ATerm arg12 = (((ATermAppl) arg1).getArity() > 1) ? ((ATermAppl) arg1).getArgument(1) : null;
			AFun arg1fun = ((ATermAppl) arg1).getAFun();
			
			if(arg1fun.equals(ATFUN)){//theorems 28, 29 [hybrid prop. metric temporal logic paper]
				norm = normalize(makeAt( arg11, arg12 ));
			} else if (arg1fun.equals(SOMEFUTFUN) && fun.equals(SOMEFUTFUN)) { //reduction of multiple somef, allf, somep, allp -  S4 modal system: somepsomep = somep, allpallp = allp
                norm = normalize(makeSomefuture(arg11));
            } else if (arg1fun.equals(SOMEPASTFUN) && fun.equals(SOMEPASTFUN)) {
                norm = normalize(makeSomepast(arg11));
            } else if (arg1fun.equals(FUTFUN) || arg1fun.equals(PASTFUN)) {
                ATerm x = (ATermAppl)normalize(factory.makeAppl(fun,nnf(arg11)));  
                norm = factory.makeAppl(arg1fun, x, arg12);
                norm = normalize(norm);
            } else if (arg1fun.equals(ALLFUTFUN) && fun.equals(SOMEFUTFUN)) {   //somef allf (if somef - normalize arg111, and remain somefallfsomef (norm(arg111)) as it was, otherwise if allf, should become somefallf(normalized arg111)
                ATerm arg111 = (((ATermAppl) arg11).getArity() > 0) ? ((ATermAppl) arg11).getArgument(0) : null;
                AFun arg11fun = ((ATermAppl) arg11).getAFun();
                if (arg11fun.equals(SOMEFUTFUN)) {
                    ATerm x = (ATermAppl)normalize((ATermAppl)ATermUtils.nnf(arg111));
                    norm = factory.makeAppl(fun,factory.makeAppl(arg1fun,factory.makeAppl(arg11fun,x)));
                }
            } else if (arg1fun.equals(ALLPASTFUN) && fun.equals(SOMEPASTFUN)) {   //somep allp (if somep - normalize arg111, and remain somepallpsomep (norm(arg111)) as it was, otherwise if allp, should become somepallp(norm(arg111))
                ATerm arg111 = (((ATermAppl) arg11).getArity() > 0) ? ((ATermAppl) arg11).getArgument(0) : null;
                AFun arg11fun = ((ATermAppl) arg11).getAFun();
                if (arg11fun.equals(SOMEPASTFUN)) {
                    ATerm x = (ATermAppl)normalize((ATermAppl)ATermUtils.nnf(arg111));
                    norm = factory.makeAppl(fun,factory.makeAppl(arg1fun,factory.makeAppl(arg11fun,x)));
                }
            }

		}
//TODO non-reducible are: allp, allpsomep, allpsomepallp, somep, somepallp, somepallpsomep and their pairs with "not"
        else if(fun.equals(ALLFUTFUN) || (fun.equals(ALLPASTFUN))) {
            ATerm arg11 = (((ATermAppl) arg1).getArity() > 0) ? ((ATermAppl) arg1).getArgument(0) : null;
            AFun arg1fun = ((ATermAppl) arg1).getAFun();
            if (arg1fun.equals(FUTFUN) || arg1fun.equals(PASTFUN)) {
                ATerm arg12 = (((ATermAppl) arg1).getArity() > 0) ? ((ATermAppl) arg1).getArgument(1) : null;
                ATerm x = (ATermAppl)normalize(factory.makeAppl(fun,nnf(arg11)));  
                norm = factory.makeAppl(arg1fun, x, arg12);
                norm = normalize(norm);
            } else if (arg1fun.equals(ALLFUTFUN) && fun.equals(ALLFUTFUN)) { //reduction of multiples
                //norm = factory.makeAppl(ALLFUTFUN,arg11);
                norm = normalize(makeAllfuture(arg11));
            } else if (arg1fun.equals(ALLPASTFUN) && fun.equals(ALLPASTFUN)) {
                //norm = factory.makeAppl(ALLFUTFUN,arg11);
                norm = normalize(makeAllpast(arg11));
            }  else if (arg1fun.equals(SOMEFUTFUN) && fun.equals(ALLFUTFUN)) {   //allf somef (allf arg111) <=>  allf somef (allf norm(arg111)), otherwise, if allf somef (somef arg111) <=> allfsomef(norm(arg111))
                ATerm arg111 = (((ATermAppl) arg11).getArity() > 0) ? ((ATermAppl) arg11).getArgument(0) : null;
                AFun arg11fun = ((ATermAppl) arg11).getAFun();
                if (arg11fun.equals(ALLFUTFUN)) {
                    ATerm x = (ATermAppl)normalize((ATermAppl)ATermUtils.nnf(arg111));
                    norm = factory.makeAppl(fun,factory.makeAppl(arg1fun,factory.makeAppl(arg11fun,x)));
                }
            } else if (arg1fun.equals(SOMEPASTFUN) && fun.equals(ALLPASTFUN)) {   //allp somep (allp arg111) <=> allp somep (allp norm(arg111)), otherwise, if allp somep (somep arg111) <=> allpsomep(norm(arg111))
                ATerm arg111 = (((ATermAppl) arg11).getArity() > 0) ? ((ATermAppl) arg11).getArgument(0) : null;
                AFun arg11fun = ((ATermAppl) arg11).getAFun();
                if (arg11fun.equals(ALLPASTFUN)) {
                    ATerm x = (ATermAppl)normalize((ATermAppl)ATermUtils.nnf(arg111));
                    norm = factory.makeAppl(fun,factory.makeAppl(arg1fun,factory.makeAppl(arg11fun,x)));
                }
            }
        }

        else if(fun.equals(ATFUN)){
            ATerm arg11 = (((ATermAppl) arg1).getArity() > 0) ? ((ATermAppl) arg1).getArgument(0) : null;
			AFun arg1fun = ((ATermAppl) arg1).getAFun();

			if(arg1fun.equals(ATFUN)){// at a at b phi <=> at b phi
				ATerm arg111 = (((ATermAppl) arg11).getArity() >0) ? ((ATermAppl) arg11).getArgument(0):null;
                ATerm arg112 = (((ATermAppl) arg11).getArity() >1) ? ((ATermAppl) arg11).getArgument(1):null;

//                i++;
                norm = normalize(makeAt( (ATerm)((ATermAppl)arg111), (ATerm)((ATermAppl)arg112)))/*nnf(arg111), arg12 )*/;
			} else if (arg1fun.equals(FUTFUN) || arg1fun.equals(PASTFUN)) {

                ATerm arg12 = (((ATermAppl) arg1).getArity() > 0) ? ((ATermAppl) arg1).getArgument(1) : null;
//
                norm = factory.makeAppl(arg1fun, normalize((ATermAppl)arg11),arg12);
                norm = normalize(norm);
                norm = factory.makeAppl(ATFUN, norm, arg2);
            } else if (arg1fun.equals(NOTFUN)) {
                norm = factory.makeAppl(arg1fun, normalize((ATermAppl)arg11));
                norm = normalize(norm);
                norm = factory.makeAppl(ATFUN,norm, arg2);
            }
            else if (((ATermAppl) arg1).getArity() > 0){//TODO cases somefut/allfut/somepast/allpast/ and just atomic concept
                ATerm xxx = normalize((ATermAppl)arg1);
                norm = factory.makeAppl(arg1fun, xxx);
                norm = normalize(norm);
                norm = factory.makeAppl(ATFUN,norm, arg2);
            }
        }
       	return norm;

	}

	/**
	 * simplfiy the term by making following changes:
	 * <ul>
	 *   <li>and([]) -> TOP</li>
	 *   <li>all(p, TOP) -> TOP</li>
	 *   <li>min(p, 0) -> TOP</li>
	 *   <li>and([a1, and([a2,...,an])]) -> and([a1, a2, ..., an]))</li>
	 *   <li>and([a, not(a), ...]) -> BOTTOM</li>
	 *   <li>not(C) -> not(simplify(C))</li>
     *   <li>and (at x (a),at x (b),at x (c),...)->at x and (a,b,c,...)</li>
	 * </ul>
	 * 
	 * @param term
	 * @return
	 */
	public static ATermAppl simplify( ATermAppl term ) {
        ATermAppl simp = term;
        AFun fun = term.getAFun();
        ATerm arg1 = (term.getArity() > 0) ? term.getArgument( 0 ) : null;
        ATerm arg2 = (term.getArity() > 1) ? term.getArgument( 1 ) : null;

        if( fun.equals( NOTFUN ) ) {
            ATermAppl arg = (ATermAppl) arg1;
            if( isNot( arg ) )
                simp = simplify( (ATermAppl) arg.getArgument( 0 ) );
            else if( isMin( arg ) ) {
                ATermInt n = (ATermInt) arg.getArgument( 1 );
                if( n.getInt() == 0 ) simp = BOTTOM;
            }
        }
        else if( fun.equals( ANDFUN ) ) {
            ATermList conjuncts = (ATermList) arg1;
            if( conjuncts.isEmpty() )
                simp = TOP;
            else {
                Set set = new HashSet();
                List negations = new ArrayList();
                MultiListIterator i = new MultiListIterator( conjuncts );
                while( i.hasNext() ) {
                    ATermAppl c = (ATermAppl) i.next();
                    if( c.equals( TOP ) )
                        continue;
                    else if( c.equals( BOTTOM ) )
                        return BOTTOM;
                    else if( isAnd( c ) )
                        i.append( (ATermList) c.getArgument( 0 ) );
                    else if( isNot( c ) )
                        negations.add( c );                    
                    else
                        set.add( c );
                }
                
                for( Iterator j = negations.iterator(); j.hasNext(); ) {
                    ATermAppl notC = (ATermAppl) j.next();
                    ATermAppl c = (ATermAppl) notC.getArgument( 0 );
                    if( set.contains( c ) )
                        return BOTTOM;
                }

                if( set.isEmpty() ) {
                    if( negations.isEmpty() )
                        return TOP;
                    else if( negations.size() == 1 )
                        return (ATermAppl) negations.get( 0 );
                }
                else if( set.size() == 1 && negations.isEmpty() )
                    return (ATermAppl) set.iterator().next();
                
                negations.addAll( set );                
                int size = negations.size();
                ATermAppl[] terms = new ATermAppl[size];
                negations.toArray( terms );
                simp = makeAnd( toSet( terms, size ) );                
            }
        }
        else if( fun.equals( ALLFUN ) ) {
            if( arg2.equals( TOP ) ) simp = TOP;
        }
        else if( fun.equals( MINFUN ) ) {
            ATermInt n = (ATermInt) arg2;
            if( n.getInt() == 0 ) simp = TOP;
        }

        return simp;
    }

	/**
	 * Creates a simplified and assuming that all the elements have already been normalized. 
	 * 
	 * @param conjuncts
	 * @return
	 */
	public static ATermAppl makeSimplifiedAnd( Collection conjuncts ) {
        Set set = new HashSet();
        List negations = new ArrayList();
        MultiListIterator listIt = new MultiListIterator( EMPTY_LIST );
        MultiIterator i = new MultiIterator( conjuncts.iterator(), listIt );
        while( i.hasNext() ) {
            ATermAppl c = (ATermAppl) i.next();
            if( c.equals( TOP ) )
                continue;
            else if( c.equals( BOTTOM ) )
                return BOTTOM;
            else if( isAnd( c ) )
                listIt.append( (ATermList) c.getArgument( 0 ) );
            else if( isNot( c ) )
                negations.add( c );                    
            else
                set.add( c );
        }
        
        for( Iterator j = negations.iterator(); j.hasNext(); ) {
            ATermAppl notC = (ATermAppl) j.next();
            ATermAppl c = (ATermAppl) notC.getArgument( 0 );
            if( set.contains( c ) )
                return BOTTOM;
        }

        if( set.isEmpty() ) {
            if( negations.isEmpty() )
                return TOP;
            else if( negations.size() == 1 )
                return (ATermAppl) negations.get( 0 );
        }
        else if( set.size() == 1 && negations.isEmpty() )
            return (ATermAppl) set.iterator().next();
        
        negations.addAll( set );                
        int size = negations.size();
        ATermAppl[] terms = new ATermAppl[size];
        negations.toArray( terms );
        return makeAnd( toSet( terms, size ) );  	    
	}
	
	/**
	 * @param ter
	 * @return
	 */
	
	//Recursive function that finds all the primitive classes
	//named in a complex concept, not looking inside restrictions
	//Added for topological sorting of quasi definition order.

    //TODO add finding primitive temporal concepts.
	public static Set findPrimitives(ATermAppl ter) {
		Set result= new HashSet();
		if (isAnd(ter) || isOr(ter) || isNot(ter)  ){
		   ATermList args = ter.getArguments();
		   
		   while(!args.isEmpty()){
		   	if (args.getFirst() instanceof ATermAppl){  	   
		   	    
		   	    result.addAll(findPrimitives( (ATermAppl)args.getFirst()));
			   	args = args.getNext();
		   		
		     }
		   	if (args.getFirst() instanceof ATermList){
			        ATermList lis =  (ATermList)args.getFirst();
			   	    while(!lis.isEmpty()){
			   	       	result.addAll(	findPrimitives((ATermAppl)lis.getFirst()));
			   	    	lis= lis.getNext();
			   	    }
			   	   args = args.getNext();
			   		
		    }
		  }
		}
		
		if (isAllValues(ter) ||isSomeValues(ter)){
			ATermList args = ter.getArguments();
			args= args.getNext(); 
			if (args.getFirst() instanceof ATermAppl){  	   
		   	    
		   	    result.addAll(findPrimitives( (ATermAppl)args.getFirst()));
			   	args = args.getNext();
		   		
		     }
		   	if (args.getFirst() instanceof ATermList){
			        ATermList lis =  (ATermList)args.getFirst();
			   	    while(!lis.isEmpty()){
			   	       	result.addAll(	findPrimitives((ATermAppl)lis.getFirst()));
			   	    	lis= lis.getNext();
			   	    }
			   	   args = args.getNext();
			   		
		    }
		  }
        if (isFuture(ter)||isPast(ter)) {
            ATermList args = ter.getArguments();
//ToDO distance (integer)
		    while(!args.isEmpty()){
		   	if (args.getFirst() instanceof ATermAppl){
		   	    result.addAll(findPrimitives( (ATermAppl)args.getFirst()));
			   	args = args.getNext();
		     }
            if (args.getFirst() instanceof ATermList){
			        ATermList lis =  (ATermList)args.getFirst();
			   	    while(!lis.isEmpty()){
			   	       	result.addAll(	findPrimitives((ATermAppl)lis.getFirst()));
			   	    	lis= lis.getNext();
			            }
            }
            if ((args.getFirst() instanceof ATerm)) {
                break;
                //  args = args.getNext();
            }
			    args = args.getNext();
		    }
            }
//ToDO add recognition of other temporal constructs

		if (isPrimitive(ter) ){
			result.add(ter);
		}
		
		
		
		
		return result;
	}
//**********************************************************

 public static void main (String[] args) {
    		ATermFactory factory = ATermUtils.getFactory();
            ATerm t1,t2,t3 = null;
            ATerm norm = null;
            ATermAppl norms = null;
//            t1 = factory.parse("or(not(at(A,5)),B))");
//            t2 = factory.parse("or(not(at(at(A,4),5)),B)");
//            t3 = factory.parse("and(not(A),A)");
//              t1 = factory.parse("and(not(allpst(not(not(C))))");
            t1 = factory.parse("and((not(somepast(not(not(C)))),past(C,1)))");
			System.out.println(t1);
            System.out.println("\n");
            System.out.println(t3);
            System.out.println("\n");

     norms = (ATermAppl)t1;
     norm = nnf(norms);
//     norm = simplify((ATermAppl) norm);
//     norm = normalize((ATermAppl)norm);
//     norm = normalize((ATermAppl)norms);
     System.out.println(norm);
 }


}//end class
